/* ----------------------------------------------------------------------------
 *         ATMEL Microcontroller Software Support  -  ROUSSET  -
 * ----------------------------------------------------------------------------
 * Copyright (c) 2006, Atmel Corporation

 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the disclaiimer below.
 *
 * - Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the disclaimer below in the documentation and/or
 * other materials provided with the distribution.
 *
 * Atmel's name may not be used to endorse or promote products derived from
 * this software without specific prior written permission.
 *
 * DISCLAIMER: THIS SOFTWARE IS PROVIDED BY ATMEL "AS IS" AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT ARE
 * DISCLAIMED. IN NO EVENT SHALL ATMEL BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
 * OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * ----------------------------------------------------------------------------
 */

/*
$Id: bot_driver.c 198 2006-10-30 10:01:09Z jjoannic $
*/

//------------------------------------------------------------------------------
//      Includes
//------------------------------------------------------------------------------

#include "core/common.h"
#include "core/device.h"
#include "core/board.h"
#include "core/trace.h"
#include "core/usb.h"
#include "core/standard.h"

#include "msd.h"
#include "media.h"
#include "sbc.h"
#include "lun.h"
#include "bot_driver.h"
#include "sbc_methods.h"

//------------------------------------------------------------------------------
//      Global variables
//------------------------------------------------------------------------------

//! \brief  Descriptors used by the BOT driver

//! \brief  Device descriptor
//! \see    S_usb_device_descriptor
const S_usb_device_descriptor sDevice = {

    sizeof(S_usb_device_descriptor), // Size of this descriptor in bytes
    USB_DEVICE_DESCRIPTOR,           // DEVICE Descriptor Type
    USB2_00,                         // USB specification 2.00
    0x00,                            // Class is specified in interface
    0x00,                            // Subclass is specified in interface
    0x00,                            // Protocol is specified in interface
    USB_ENDPOINT0_MAXPACKETSIZE,     // Maximum packet size for endpoint zero
    USB_VENDOR_ATMEL,                // Vendor ID (ATMEL, 03EBh)
    BOT_PRODUCT_ID,                  // Product ID (6202h)
    0x0001,                          // Device release number is 0.01
    0x01,                            // Index of manufacturer string
    0x02,                            // Index of product string
    0x03,                            // Index of serial number string
    0x01                             // One possible configuration
};

//! \brief  Configuration descriptor
//! \see    S_bot_configuration_descriptor
//! \see    S_usb_configuration_descriptor
//! \see    S_usb_interface_descriptor
//! \see    S_usb_endpoint_descriptor
const S_bot_configuration_descriptor sConfiguration = {

    // Configuration Descriptor
    {
        sizeof(S_usb_configuration_descriptor), // Size of this descriptor
        USB_CONFIGURATION_DESCRIPTOR,           // CONFIGURATION descriptor type
        sizeof(S_bot_configuration_descriptor), // Total size of descriptors
        0x01,                                   // One interface
        0x01,                                   // Configuration number 1
        0x00,                                   // No string description
        USB_CONFIG_SELF_NOWAKEUP,               // Device is self-powered
                                                // Remote wakeup not supported
        USB_POWER_MA(100),                      // 100mA consumption
    },
    // MSD Class Interface Descriptor
    {
        sizeof(S_usb_interface_descriptor), // Size of this descriptor in bytes
        USB_INTERFACE_DESCRIPTOR,           // INTERFACE descriptor type
        0x00,                               // Interface number 0
        0x00,                               // Setting 0
        BOT_NUM_ENDPOINTS - 1,              // Two endpoints used (excluding endpoint 0)
        USB_CLASS_MASS_STORAGE,             // Mass storage class code
        MSD_SUBCLASS_SCSI,                  // SCSI subclass code
        MSD_PROTOCOL_BULK_ONLY,             // Bulk-only transport protocol
        0x00                                // No string description
    },
    // Bulk-OUT Endpoint Descriptor
    {
        sizeof(S_usb_endpoint_descriptor),   // Size of this descriptor in bytes
        USB_ENDPOINT_DESCRIPTOR,             // ENDPOINT descriptor type
        USB_ENDPOINT_OUT | BOT_EPT_BULK_OUT, // OUT endpoint, address 01h
        ENDPOINT_TYPE_BULK,                  // Bulk endpoint
        64,                                  // Maximum packet size is 64 bytes
        0x00,                                // Must be 0 for full-speed bulk
    },
    // Bulk_IN Endpoint Descriptor
    {
        sizeof(S_usb_endpoint_descriptor), // Size of this descriptor in bytes
        USB_ENDPOINT_DESCRIPTOR,           // ENDPOINT descriptor type
        USB_ENDPOINT_IN | BOT_EPT_BULK_IN, // IN endpoint, address 02h
        ENDPOINT_TYPE_BULK,                // Bulk endpoint
        64,                                // Maximum packet size if 64 bytes
        0x00,                              // Must be 0 for full-speed bulk
    }
};

//! \brief  Language ID
//! \see    S_usb_language_id
static const S_usb_language_id sLanguageID = {

    USB_STRING_DESCRIPTOR_SIZE(1),
    USB_STRING_DESCRIPTOR,
    USB_LANGUAGE_ENGLISH_US
};

//! \brief  Manufacturer description string
static const char pStringManufacturer[] = {

    USB_STRING_DESCRIPTOR_SIZE(5),
    USB_STRING_DESCRIPTOR,
    USB_UNICODE('A'),
    USB_UNICODE('T'),
    USB_UNICODE('M'),
    USB_UNICODE('E'),
    USB_UNICODE('L')
};

//! \brief  Product description string
static const char pStringProduct[] = {

    USB_STRING_DESCRIPTOR_SIZE(14),
    USB_STRING_DESCRIPTOR,
    USB_UNICODE('A'),
    USB_UNICODE('T'),
    USB_UNICODE('M'),
    USB_UNICODE('E'),
    USB_UNICODE('L'),
    USB_UNICODE(' '),
    USB_UNICODE('A'),
    USB_UNICODE('T'),
    USB_UNICODE('9'),
    USB_UNICODE('1'),
    USB_UNICODE(' '),
    USB_UNICODE('M'),
    USB_UNICODE('S'),
    USB_UNICODE('D')
};

//! \brief  Device serial number
//!
//!         The serial number must be at least 12 characters long and made up
//!         of only letters & numbers to be compliant with the MSD specification
static const char pStringSerialNumber[] = {

    USB_STRING_DESCRIPTOR_SIZE(12),
    USB_STRING_DESCRIPTOR,
    USB_UNICODE('0'),
    USB_UNICODE('1'),
    USB_UNICODE('2'),
    USB_UNICODE('3'),
    USB_UNICODE('4'),
    USB_UNICODE('5'),
    USB_UNICODE('6'),
    USB_UNICODE('7'),
    USB_UNICODE('8'),
    USB_UNICODE('9'),
    USB_UNICODE('A'),
    USB_UNICODE('B')
};

//! \brief  List of string descriptors used by the device
static const char *pStrings[] = {

    (char *) &sLanguageID,
    pStringManufacturer,
    pStringProduct,
    pStringSerialNumber
};

//! \brief  List of endpoint descriptors used by the device
static const S_usb_endpoint_descriptor *pEndpoints[] = {

    &(sConfiguration.sBulkOut),
    &(sConfiguration.sBulkIn)
};

//! \brief  List of standard descriptors
//! \see    S_std_descriptors
static const S_std_descriptors sDescriptors = {

    &sDevice,
    (S_usb_configuration_descriptor *) &sConfiguration,
    pStrings,
    pEndpoints
};

//------------------------------------------------------------------------------
//      Internal functions
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
//! \brief  Returns the expected transfer length and direction (IN, OUT or don't
//!         care) from the host point-of-view.
//! \param  pCbw    Pointer to the CBW to examinate
//! \param  pLength Expected length of command
//! \param  pType   Expected direction of command
//------------------------------------------------------------------------------
static void BOT_GetCommandInformation(S_msd_cbw     *pCbw,
                                      unsigned int  *pLength,
                                      unsigned char *pType)
{
    // Expected host transfer direction and length
    (*pLength) = pCbw->dCBWDataTransferLength;

    if (*pLength == 0) {

        (*pType) = BOT_NO_TRANSFER;
    }
    else if (ISSET(pCbw->bmCBWFlags, MSD_CBW_DEVICE_TO_HOST)) {

        (*pType) = BOT_DEVICE_TO_HOST;
    }
    else {

        (*pType) = BOT_HOST_TO_DEVICE;
    }
}

//------------------------------------------------------------------------------
//! \brief  Pre-processes a command by checking the differences between the
//!         host and device expectations in term of transfer type and length.
//!
//!         Once one of the thirteen cases is identified, the actions to do
//!         during the post-processing phase are stored in the dCase variable
//!         of the command state.
//! \param  pBot Pointer to a S_bot instance
//! \return true if the command is supported, false otherwise
//------------------------------------------------------------------------------
static bool BOT_PreProcessCommand(S_bot *pBot)
{
    unsigned int        dHostLength;
    unsigned int        dDeviceLength;
    unsigned char       bHostType;
    unsigned char       bDeviceType;
    bool                isCommandSupported;
    S_bot_command_state *pCommandState = &(pBot->sCommandState);
    S_msd_csw           *pCsw = &(pCommandState->sCsw);
    S_msd_cbw           *pCbw = &(pCommandState->sCbw);
    S_lun               *pLun = &(pBot->pLun[(unsigned char) pCbw->bCBWLUN]);

    // Get information about the command
    // Host-side
    BOT_GetCommandInformation(pCbw, &dHostLength, &bHostType);

    // Device-side
    isCommandSupported = SBC_GetCommandInformation(pCbw->pCommand,
                                                   &dDeviceLength,
                                                   &bDeviceType,
                                                   pLun);

    // Initialize data residue and result status
    pCsw->dCSWDataResidue = 0;
    pCsw->bCSWStatus = MSD_CSW_COMMAND_PASSED;

    // Check if the command is supported
    if (isCommandSupported) {

        // Identify the command case
        // Case 1  (Hn = Dn)
        if ((bHostType == BOT_NO_TRANSFER)
            && (bDeviceType == BOT_NO_TRANSFER)) {

            pCommandState->bCase = 0;
            pCommandState->dLength = 0;
        }
        // Case 2  (Hn < Di)
        else if ((bHostType == BOT_NO_TRANSFER)
                 && (bDeviceType == BOT_DEVICE_TO_HOST)) {

            TRACE_WARNING("W: BOT_PreProcessCommand: Case 2\n\r");
            pCommandState->bCase = BOT_CASE_PHASE_ERROR;
            pCommandState->dLength = 0;
        }
        // Case 3  (Hn < Do)
        else if ((bHostType == BOT_NO_TRANSFER)
                 && (bDeviceType == BOT_HOST_TO_DEVICE)) {

            TRACE_WARNING("W: BOT_PreProcessCommand: Case 3\n\r");
            pCommandState->bCase = BOT_CASE_PHASE_ERROR;
            pCommandState->dLength = 0;
        }
        // Case 4  (Hi > Dn)
        else if ((bHostType == BOT_DEVICE_TO_HOST)
                 && (bDeviceType == BOT_NO_TRANSFER)) {

            TRACE_WARNING("W: BOT_PreProcessCommand: Case 4\n\r");
            pCommandState->bCase = BOT_CASE_STALL_IN;
            pCommandState->dLength = 0;
            pCsw->dCSWDataResidue = dHostLength;
        }
        // Case 5  (Hi > Di)
        else if ((bHostType == BOT_DEVICE_TO_HOST)
                 && (bDeviceType == BOT_DEVICE_TO_HOST)
                 && (dHostLength > dDeviceLength)) {

            TRACE_WARNING("W: BOT_PreProcessCommand: Case 5\n\r");
            pCommandState->bCase = BOT_CASE_STALL_IN;
            pCommandState->dLength = dDeviceLength;
            pCsw->dCSWDataResidue = dHostLength - dDeviceLength;
        }
        // Case 6  (Hi = Di)
        if ((bHostType == BOT_DEVICE_TO_HOST)
            && (bDeviceType == BOT_DEVICE_TO_HOST)
            && (dHostLength == dDeviceLength)) {

            pCommandState->bCase = 0;
            pCommandState->dLength = dDeviceLength;
        }
        // Case 7  (Hi < Di)
        else if ((bHostType == BOT_DEVICE_TO_HOST)
                 && (bDeviceType == BOT_DEVICE_TO_HOST)
                 && (dHostLength < dDeviceLength)) {

            TRACE_WARNING("W: BOT_PreProcessCommand: Case 7\n\r");
            pCommandState->bCase = BOT_CASE_PHASE_ERROR;
            pCommandState->dLength = dHostLength;
        }
        // Case 8  (Hi <> Do)
        else if ((bHostType == BOT_DEVICE_TO_HOST)
                 && (bDeviceType == BOT_HOST_TO_DEVICE)) {

            TRACE_WARNING("W: BOT_PreProcessCommand: Case 8\n\r");
            pCommandState->bCase = BOT_CASE_STALL_IN | BOT_CASE_PHASE_ERROR;
            pCommandState->dLength = 0;
        }
        // Case 9  (Ho > Dn)
        else if ((bHostType == BOT_HOST_TO_DEVICE)
                 && (bDeviceType == BOT_NO_TRANSFER)) {

            TRACE_WARNING("W: BOT_PreProcessCommand: Case 9\n\r");
            pCommandState->bCase = BOT_CASE_STALL_OUT;
            pCommandState->dLength = 0;
            pCsw->dCSWDataResidue = dHostLength;
        }
        // Case 10 (Ho <> Di)
        else if ((bHostType == BOT_HOST_TO_DEVICE)
                 && (bDeviceType == BOT_DEVICE_TO_HOST)) {

            TRACE_WARNING("W: BOT_PreProcessCommand: Case 10\n\r");
            pCommandState->bCase = BOT_CASE_STALL_OUT | BOT_CASE_PHASE_ERROR;
            pCommandState->dLength = 0;
        }
        // Case 11 (Ho > Do)
        else if ((bHostType == BOT_HOST_TO_DEVICE)
                 && (bDeviceType == BOT_HOST_TO_DEVICE)
                 && (dHostLength > dDeviceLength)) {

            TRACE_WARNING("W: BOT_PreProcessCommand: Case 11\n\r");
            pCommandState->bCase = BOT_CASE_STALL_OUT;
            pCommandState->dLength = dDeviceLength;
            pCsw->dCSWDataResidue = dHostLength - dDeviceLength;
        }
        // Case 12 (Ho = Do)
        else if ((bHostType == BOT_HOST_TO_DEVICE)
                 && (bDeviceType == BOT_HOST_TO_DEVICE)
                 && (dHostLength == dDeviceLength)) {

            pCommandState->bCase = 0;
            pCommandState->dLength = dDeviceLength;
        }
        // Case 13 (Ho < Do)
        else if ((bHostType == BOT_HOST_TO_DEVICE)
                 && (bDeviceType == BOT_HOST_TO_DEVICE)
                 && (dHostLength < dDeviceLength)) {

            TRACE_WARNING("W: BOT_PreProcessCommand: Case 13\n\r");
            pCommandState->bCase = BOT_CASE_PHASE_ERROR;
            pCommandState->dLength = dHostLength;
        }
    }

    return isCommandSupported;
}

//------------------------------------------------------------------------------
//! \brief  Post-processes a command given the case identified during the
//!         pre-processing step.
//!
//!         Depending on the case, one of the following actions can be done:
//!             - Bulk IN endpoint is stalled
//!             - Bulk OUT endpoint is stalled
//!             - CSW status set to phase error
//! \param  pBot Pointer to a S_bot instance
//------------------------------------------------------------------------------
static void BOT_PostProcessCommand(S_bot *pBot)
{
    S_bot_command_state *pCommandState = &(pBot->sCommandState);
    S_msd_csw           *pCsw = &(pCommandState->sCsw);
    const S_usb         *pUsb = pBot->sClass.pUsb;

    // STALL Bulk IN endpoint ?
    if (ISSET(pCommandState->bCase, BOT_CASE_STALL_IN)) {

        TRACE_DEBUG_M("StallIn ");
        USB_Halt(pUsb, BOT_EPT_BULK_IN, USB_SET_FEATURE);
    }

    // STALL Bulk OUT endpoint ?
    if (ISSET(pCommandState->bCase, BOT_CASE_STALL_OUT)) {

        TRACE_DEBUG_M("StallOut ");
        USB_Halt(pUsb, BOT_EPT_BULK_OUT, USB_SET_FEATURE);
    }

    // Set CSW status code to phase error ?
    if (ISSET(pCommandState->bCase, BOT_CASE_PHASE_ERROR)) {

        TRACE_DEBUG_M("PhaseErr ");
        pCsw->bCSWStatus = MSD_CSW_PHASE_ERROR;
    }
}

//------------------------------------------------------------------------------
//! \brief  Processes the latest command received by the device.
//! \param  pBot Pointer to a S_bot instance
//! \return true if the command has been completed, false otherwise.
//------------------------------------------------------------------------------
static bool BOT_ProcessCommand(S_bot *pBot)
{
    unsigned char       bStatus;
    const S_usb         *pUsb = pBot->sClass.pUsb;
    S_bot_command_state *pCommandState = &(pBot->sCommandState);
    S_msd_cbw           *pCbw = &(pCommandState->sCbw);
    S_msd_csw           *pCsw = &(pCommandState->sCsw);
    S_lun               *pLun = &(pBot->pLun[(unsigned char) pCbw->bCBWLUN]);
    bool                isCommandComplete = false;

    // Check if LUN is valid
    if (pCbw->bCBWLUN > pBot->bMaxLun) {

        TRACE_WARNING("W: BOT_ProcessCommand: Requested LUN does not exist\n\r");
        bStatus = BOT_STATUS_ERROR;
    }
    else {

        // Process command
        if (pBot->bMaxLun > 0) {

            TRACE_DEBUG_M("LUN%d ", pCbw->bCBWLUN);
        }

        bStatus = SBC_ProcessCommand(pUsb, pLun, pCommandState);
    }

    // Check command result code
    if (bStatus == BOT_STATUS_PARAMETER) {

        TRACE_WARNING("W: BOT_ProcessCommand: Unknown command 0x%02X\n\r",
                      pCbw->pCommand[0]);

        // Update sense data
        SBC_UpdateSenseData(&(pLun->sRequestSenseData),
                            SBC_SENSE_KEY_ILLEGAL_REQUEST,
                            SBC_ASC_INVALID_FIELD_IN_CDB,
                            0);

        // Result codes
        pCsw->bCSWStatus = MSD_CSW_COMMAND_FAILED;
        isCommandComplete = true;

        // stall the request, IN or OUT
        if (!ISSET(pCbw->bmCBWFlags, MSD_CBW_DEVICE_TO_HOST)
            && (pCbw->dCBWDataTransferLength > 0)) {

            // Stall the OUT endpoint : host to device
            USB_Halt(pUsb, BOT_EPT_BULK_OUT, USB_SET_FEATURE);
            TRACE_DEBUG_M("StaOUT ");
        }
        else {

            // Stall the IN endpoint : device to host
            USB_Halt(pUsb, BOT_EPT_BULK_IN, USB_SET_FEATURE);
            TRACE_DEBUG_M("StaIN ");
        }
    }
    else if (bStatus == BOT_STATUS_ERROR) {

        TRACE_WARNING("W: MSD_ProcessCommand: Command failed\n\r");

        // Update sense data
// TODO (jjoannic#1#): Change code
        SBC_UpdateSenseData(&(pLun->sRequestSenseData),
                            SBC_SENSE_KEY_MEDIUM_ERROR,
                            SBC_ASC_INVALID_FIELD_IN_CDB,
                            0);

        // Result codes
        pCsw->bCSWStatus = MSD_CSW_COMMAND_FAILED;
        isCommandComplete = true;
    }
    else {

        // Update sense data
        SBC_UpdateSenseData(&(pLun->sRequestSenseData),
                            SBC_SENSE_KEY_NO_SENSE,
                            0,
                            0);

        // Is command complete ?
        if (bStatus == BOT_STATUS_SUCCESS) {

            isCommandComplete = true;
        }
    }

    // Check if command has been completed
    if (isCommandComplete) {

        TRACE_DEBUG_M("Cplt ");

        // Adjust data residue
        if (pCommandState->dLength != 0) {

            pCsw->dCSWDataResidue += pCommandState->dLength;

            // STALL the endpoint waiting for data
            if (!ISSET(pCbw->bmCBWFlags, MSD_CBW_DEVICE_TO_HOST)) {

                // Stall the OUT endpoint : host to device
                USB_Halt(pUsb, BOT_EPT_BULK_OUT, USB_SET_FEATURE);
                TRACE_DEBUG_M("StaOUT ");
            }
            else {

                // Stall the IN endpoint : device to host
                USB_Halt(pUsb, BOT_EPT_BULK_IN, USB_SET_FEATURE);
                TRACE_DEBUG_M("StaIN ");
            }
        }

        // Reset command state
        pCommandState->bState = 0;
    }

    return isCommandComplete;
}

//------------------------------------------------------------------------------
//      Exported functions
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
//! \brief  Resets the state of the BOT driver
//! \param  pBot Pointer to a S_bot instance
//! \see    S_bot
//------------------------------------------------------------------------------
void BOT_Reset(S_bot *pBot)
{
    TRACE_DEBUG_M("MSDReset ");

#if !defined(TR_DEBUG_L)
    TRACE_DEBUG_M("\n\r");
#endif

    pBot->bState = BOT_STATE_READ_CBW;
    pBot->isWaitResetRecovery = false;
    pBot->sCommandState.bState = 0;
}

//------------------------------------------------------------------------------
//! \brief  Initializes a BOT driver and the associated USB driver.
//! \param  pBot    Pointer to a S_bot instance
//! \param  pUsb    USB driver to use
//! \param  pLun    Pointer to a list of LUNs
//! \param  bNumLun Number of LUN in list
//! \see    S_bot
//! \see    S_usb
//------------------------------------------------------------------------------
void BOT_Init(S_bot *pBot,
              const S_usb *pUsb,
              S_lun *pLun,
              unsigned char bNumLun)
{
    TRACE_INFO("I: MSD init\n\r");

    // BOT driver initialization
    pBot->sClass.pUsb = pUsb;
    pBot->sClass.pDescriptors = &sDescriptors;

    // Command state initialization
    pBot->sCommandState.bState = 0;
    pBot->sCommandState.bCase = 0;
    pBot->sCommandState.dLength = 0;
    pBot->sCommandState.sTransfer.bSemaphore = 0;

    // LUNs
    pBot->pLun = pLun;
    pBot->bMaxLun = (unsigned char) (bNumLun - 1);

    // Reset BOT driver
    BOT_Reset(pBot);

    // Init the USB driver
    USB_Init(pUsb);
}

//------------------------------------------------------------------------------
//! \brief  Handler for incoming SETUP requests on default Control endpoint 0.
//!
//!         Standard requests are forwarded to the STD_RequestHandler method.
//! \param  pBot Pointer to a S_bot instance
//------------------------------------------------------------------------------
void BOT_RequestHandler(S_bot *pBot)
{
    const S_usb *pUsb = pBot->sClass.pUsb;
    S_usb_request *pSetup = USB_GetSetup(pUsb);

    TRACE_DEBUG_M("NewReq ");

    // Handle requests
    switch (pSetup->bRequest) {
    //---------------------
    case USB_CLEAR_FEATURE:
    //---------------------
        TRACE_DEBUG_M("ClrFeat ");

        switch (pSetup->wValue) {

        //---------------------
        case USB_ENDPOINT_HALT:
        //---------------------
            TRACE_DEBUG_M("Hlt ");

            // Do not clear the endpoint halt status if the device is waiting
            // for a reset recovery sequence
            if (!pBot->isWaitResetRecovery) {

                // Forward the request to the standard handler
                STD_RequestHandler(&(pBot->sClass));
            }
            else {

                TRACE_DEBUG_M("No ");
            }

            USB_SendZLP0(pUsb, 0, 0);
            break;

        //------
        default:
        //------
            // Forward the request to the standard handler
            STD_RequestHandler(&(pBot->sClass));
        }
        break;

    //-------------------
    case MSD_GET_MAX_LUN:
    //-------------------
        TRACE_DEBUG_M("gMaxLun ");

        // Check request parameters
        if ((pSetup->wValue == 0)
            && (pSetup->wIndex == 0)
            && (pSetup->wLength == 1)) {

            USB_Write(pUsb, 0, &(pBot->bMaxLun), 1, 0, 0);
        }
        else {

            TRACE_WARNING("W: BOT_RequestHandler: GetMaxLUN(%d,%d,%d)\n\r",
                          pSetup->wValue, pSetup->wIndex, pSetup->wLength);
            USB_Stall(pUsb, 0);
        }
        break;

    //-----------------------
    case MSD_BULK_ONLY_RESET:
    //-----------------------
        TRACE_DEBUG_M("Rst ");

        // Check parameters
        if ((pSetup->wValue == 0)
            && (pSetup->wIndex == 0)
            && (pSetup->wLength == 0)) {

            // Reset the MSD driver
            BOT_Reset(pBot);
            USB_SendZLP0(pUsb, 0, 0);
        }
        else {

            TRACE_WARNING("W: BOT_RequestHandler: Reset(%d,%d,%d)\n\r",
                          pSetup->wValue, pSetup->wIndex, pSetup->wLength);
            USB_Stall(pUsb, 0);
        }
        break;

    //------
    default:
    //------
        // Forward request to standard handler
        STD_RequestHandler(&(pBot->sClass));

        break;
    }

#if !defined(TR_DEBUG_L)
    TRACE_DEBUG_M("\n\r");
#endif
}

//------------------------------------------------------------------------------
//! \brief  State machine for the BOT driver
//! \param  pBot Pointer to a S_bot instance
//------------------------------------------------------------------------------
void BOT_StateMachine(S_bot *pBot)
{
    const S_usb         *pUsb = pBot->sClass.pUsb;
    S_bot_command_state *pCommandState = &(pBot->sCommandState);
    S_msd_cbw           *pCbw = &(pCommandState->sCbw);
    S_msd_csw           *pCsw = &(pCommandState->sCsw);
    S_bot_transfer      *pTransfer = &(pCommandState->sTransfer);
    unsigned char       bStatus;

    // Identify current driver state
    switch (pBot->bState) {
    //----------------------
    case BOT_STATE_READ_CBW:
    //----------------------
        // Start the CBW read operation
        pTransfer->bSemaphore = 0;
        bStatus = USB_Read(pUsb,
                           BOT_EPT_BULK_OUT,
                           pCbw,
                           MSD_CBW_SIZE,
                           (Callback_f) BOT_Callback,
                           (void *) pTransfer);

        // Check operation result code
        if (bStatus == USB_STATUS_SUCCESS) {

            // If the command was successful, wait for transfer
            pBot->bState = BOT_STATE_WAIT_CBW;
        }
        break;

    //----------------------
    case BOT_STATE_WAIT_CBW:
    //----------------------
        // Check transfer semaphore
        if (pTransfer->bSemaphore > 0) {

            // Take semaphore and terminate transfer
            pTransfer->bSemaphore--;

            // Check if transfer was successful
            if (pTransfer->bStatus == USB_STATUS_SUCCESS) {

#if !defined(TR_DEBUG_L)
                TRACE_DEBUG_M("\n\r");
#endif

                TRACE_DEBUG_M("------------------------------\n\r");

                // Process received command
                pBot->bState = BOT_STATE_PROCESS_CBW;
            }
            else if (pTransfer->bStatus == USB_STATUS_RESET) {

                TRACE_INFO("I: BOT_StateMachine: Endpoint resetted\n\r");
                pBot->bState = BOT_STATE_READ_CBW;
            }
            else {

                TRACE_WARNING("W: BOT_StateMachine: Failed to read CBW\n\r");
                pBot->bState = BOT_STATE_READ_CBW;
            }
        }
        break;

    //-------------------------
    case BOT_STATE_PROCESS_CBW:
    //-------------------------
        // Check if this is a new command
        if (pCommandState->bState == 0) {

            // Copy the CBW tag
            pCsw->dCSWTag = pCbw->dCBWTag;

            // Check that the CBW is 31 bytes long
            if ((pTransfer->dBytesTransferred != MSD_CBW_SIZE) ||
                (pTransfer->dBytesRemaining != 0)) {

                TRACE_WARNING("W: BOT_StateMachine: Invalid CBW (too short or too long)\n\r");

                // Wait for a reset recovery
                pBot->isWaitResetRecovery = true;

                // Halt the Bulk-IN and Bulk-OUT pipes
                USB_Halt(pUsb, BOT_EPT_BULK_OUT, USB_SET_FEATURE);
                USB_Halt(pUsb, BOT_EPT_BULK_IN, USB_SET_FEATURE);

                pCsw->bCSWStatus = MSD_CSW_COMMAND_FAILED;
                pBot->bState = BOT_STATE_READ_CBW;
            }
            // Check the CBW Signature
            else if (pCbw->dCBWSignature != MSD_CBW_SIGNATURE) {

                TRACE_WARNING("W: MSD_BOTStateMachine: Invalid CBW (Bad signature)\n\r");

                // Wait for a reset recovery
                pBot->isWaitResetRecovery = true;

                // Halt the Bulk-IN and Bulk-OUT pipes
                USB_Halt(pUsb, BOT_EPT_BULK_OUT, USB_SET_FEATURE);
                USB_Halt(pUsb, BOT_EPT_BULK_IN, USB_SET_FEATURE);

                pCsw->bCSWStatus = MSD_CSW_COMMAND_FAILED;
                pBot->bState = BOT_STATE_READ_CBW;
            }
            else {

                // Pre-process command
                BOT_PreProcessCommand(pBot);
            }
        }

        // Process command
        if (pCsw->bCSWStatus == BOT_STATUS_SUCCESS) {

            if (BOT_ProcessCommand(pBot)) {

                // Post-process command if it is finished
                BOT_PostProcessCommand(pBot);
                pBot->bState = BOT_STATE_SEND_CSW;
            }
            TRACE_DEBUG_M("\n\r");
        }

        break;

    //----------------------
    case BOT_STATE_SEND_CSW:
    //----------------------
        // Set signature
        pCsw->dCSWSignature = MSD_CSW_SIGNATURE;

        // Start the CSW write operation
        bStatus = USB_Write(pUsb,
                            BOT_EPT_BULK_IN,
                            pCsw,
                            MSD_CSW_SIZE,
                            (Callback_f) BOT_Callback,
                            (void *) pTransfer);

        // Check operation result code
        if (bStatus == USB_STATUS_SUCCESS) {

            TRACE_DEBUG_M("SendCSW ");

            // Wait for end of transfer
            pBot->bState = BOT_STATE_WAIT_CSW;
        }
        break;

    //----------------------
    case BOT_STATE_WAIT_CSW:
    //----------------------
        // Check transfer semaphore
        if (pTransfer->bSemaphore > 0) {

            // Take semaphore and terminate transfer
            pTransfer->bSemaphore--;

            // Check if transfer was successful
            if (bStatus == USB_STATUS_RESET) {

                TRACE_INFO("I: BOT_StateMachine: Endpoint resetted\n\r");
            }
            else if (bStatus == USB_STATUS_ABORTED) {

                TRACE_WARNING("W: BOT_StateMachine: Failed to send CSW\n\r");
            }
            else {

                TRACE_DEBUG_M("ok");
            }

            // Read new CBW
            pBot->bState = BOT_STATE_READ_CBW;
        }
        break;
    }
}
