/* ----------------------------------------------------------------------------
 *         ATMEL Microcontroller Software Support  -  ROUSSET  -
 * ----------------------------------------------------------------------------
 * Copyright (c) 2006, Atmel Corporation

 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the disclaiimer below.
 *
 * - Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the disclaimer below in the documentation and/or
 * other materials provided with the distribution.
 *
 * Atmel's name may not be used to endorse or promote products derived from
 * this software without specific prior written permission.
 *
 * DISCLAIMER: THIS SOFTWARE IS PROVIDED BY ATMEL "AS IS" AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT ARE
 * DISCLAIMED. IN NO EVENT SHALL ATMEL BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
 * OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * ----------------------------------------------------------------------------
 */

/*
$Id: sbc_methods.c 198 2006-10-30 10:01:09Z jjoannic $
*/

//------------------------------------------------------------------------------
//      Includes
//------------------------------------------------------------------------------

#include "core/common.h"
#include "core/device.h"
#include "core/board.h"
#include "core/trace.h"
#include "core/usb.h"
#include "core/standard.h"

#include "msd.h"
#include "media.h"
#include "sbc.h"
#include "lun.h"
#include "bot_driver.h"
#include "sbc_methods.h"

//------------------------------------------------------------------------------
//      Global variables
//------------------------------------------------------------------------------

//! \brief  Header for the mode pages data
//! \see    S_sbc_mode_parameter_header_6
static const S_sbc_mode_parameter_header_6 sModeParameterHeader6 = {

    sizeof(S_sbc_mode_parameter_header_6) - 1,  //!< Length of mode page data is 0x03
    SBC_MEDIUM_TYPE_DIRECT_ACCESS_BLOCK_DEVICE, //!< Direct-access block device
    0,                                          //!< Reserved bits
    false,                                      //!< DPO/FUA not supported
    0,                                          //!< Reserved bits
    false,                                      //!< Medium is not write-protected
    0                                           //!< No block descriptor
};

//------------------------------------------------------------------------------
//      Internal functions
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
//! \brief  Performs a WRITE (10) command on the specified LUN.
//!
//!         The data to write is first received from the USB host and then
//!         actually written on the media.
//!         This function operates asynchronously and must be called multiple
//!         times to complete. A result code of BOT_STATUS_INCOMPLETE indicates
//!         that at least another call of the method is necessary.
//! \param  pUsb          Pointer to a S_usb instance
//! \param  pLun          Pointer to the LUN affected by the command
//! \param  pCommandState Current state of the command
//! \return Operation result code (SUCCESS, ERROR, INCOMPLETE or PARAMETER)
//! \see    S_usb
//! \see    S_lun
//! \see    S_bot_command_state
//------------------------------------------------------------------------------
static unsigned char SBC_Write10(const S_usb         *pUsb,
                                 S_lun               *pLun,
                                 S_bot_command_state *pCommandState)
{
    unsigned char  bStatus;
    unsigned char  bResult = BOT_STATUS_INCOMPLETE;
    S_bot_transfer *pTransfer = &(pCommandState->sTransfer);
    S_sbc_write_10 *pCommand = (S_sbc_write_10 *) pCommandState->sCbw.pCommand;

    // Init command state
    if (pCommandState->bState == 0) {

        pCommandState->bState = SBC_STATE_READ;
    }

    // Convert pLength from bytes to blocks
    pCommandState->dLength /= pLun->dBlockSize;

    // Check if pLength equals 0
    if (pCommandState->dLength == 0) {

        TRACE_DEBUG_M("End ");
        bResult = BOT_STATUS_SUCCESS;
    }
    else {

        // Current command status
        switch (pCommandState->bState) {
        //------------------
        case SBC_STATE_READ:
        //------------------
            TRACE_DEBUG_M("Receive ");
            // Read one block of data sent by the host
            bStatus = USB_Read(pUsb,
                               BOT_EPT_BULK_OUT,
                               pLun->pReadWriteBuffer,
                               pLun->dBlockSize,
                               (Callback_f) BOT_Callback,
                               (void *) pTransfer);

            // Check operation result code
            if (bStatus != USB_STATUS_SUCCESS) {

                TRACE_WARNING("W: RBC_Write10: Failed to start receiving data\n\r");
                SBC_UpdateSenseData(&(pLun->sRequestSenseData),
                                    SBC_SENSE_KEY_HARDWARE_ERROR,
                                    0,
                                    0);
                bResult = BOT_STATUS_ERROR;
            }
            else {

                // Prepare next device state
                pCommandState->bState = SBC_STATE_WAIT_READ;
            }
            break;

        //-----------------------
        case SBC_STATE_WAIT_READ:
        //-----------------------
            TRACE_DEBUG_M("Wait ");

            // Check semaphore
            if (pTransfer->bSemaphore > 0) {

                pTransfer->bSemaphore--;
                pCommandState->bState = SBC_STATE_WRITE;
            }
            break;

        //-------------------
        case SBC_STATE_WRITE:
        //-------------------
            // Check the result code of the read operation
            if (pTransfer->bStatus != USB_STATUS_SUCCESS) {

                TRACE_WARNING("W: RBC_Write10: Failed to received data\n\r");
                SBC_UpdateSenseData(&(pLun->sRequestSenseData),
                                    SBC_SENSE_KEY_HARDWARE_ERROR,
                                    0,
                                    0);
                bResult = BOT_STATUS_ERROR;
            }
            else {

                // Write the block to the media
                bStatus = LUN_Write(pLun,
                                    DWORDB(pCommand->pLogicalBlockAddress),
                                    pLun->pReadWriteBuffer,
                                    1,
                                    (Callback_f) BOT_Callback,
                                    (void *) pTransfer);

                // Check operation result code
                if (bStatus != USB_STATUS_SUCCESS) {

                    TRACE_WARNING("W: RBC_Write10: Failed to start media write\n\r");
                    SBC_UpdateSenseData(&(pLun->sRequestSenseData),
                                        SBC_SENSE_KEY_NOT_READY,
                                        0,
                                        0);
                    bResult = BOT_STATUS_ERROR;
                }
                else {

                    // Prepare next state
                    pCommandState->bState = SBC_STATE_WAIT_WRITE;
                }
            }
            break;

        //------------------------
        case SBC_STATE_WAIT_WRITE:
        //------------------------
            TRACE_DEBUG_M("Wait ");

            // Check semaphore value
            if (pTransfer->bSemaphore > 0) {

                // Take semaphore and move to next state
                pTransfer->bSemaphore--;
                pCommandState->bState = SBC_STATE_NEXT_BLOCK;
            }
            break;

        //------------------------
        case SBC_STATE_NEXT_BLOCK:
        //------------------------
            // Check operation result code
            if (pTransfer->bStatus != USB_STATUS_SUCCESS) {

                TRACE_WARNING("W: RBC_Write10: Failed to write media\n\r");
                SBC_UpdateSenseData(&(pLun->sRequestSenseData),
                                    SBC_SENSE_KEY_RECOVERED_ERROR,
                                    SBC_ASC_TOO_MUCH_WRITE_DATA,
                                    0);
                bResult = BOT_STATUS_ERROR;
            }
            else {

                // Update transfer length and block address
                pCommandState->dLength--;
                STORE_DWORDB(DWORDB(pCommand->pLogicalBlockAddress) + 1,
                             pCommand->pLogicalBlockAddress);

                // Check if transfer is finished
                if (pCommandState->dLength == 0) {

                    bResult = BOT_STATUS_SUCCESS;
                }
                else {

                    pCommandState->bState = SBC_STATE_READ;
                }
            }
            break;
        }
    }

    // Convert dLength from blocks to bytes
    pCommandState->dLength *= pLun->dBlockSize;

    return bResult;
}

//------------------------------------------------------------------------------
//! \brief  Performs a READ (10) command on specified LUN.
//!
//!         The data is first read from the media and then sent to the USB host.
//!         This function operates asynchronously and must be called multiple
//!         times to complete. A result code of BOT_STATUS_INCOMPLETE indicates
//!         that at least another call of the method is necessary.
//! \param  pUsb          Pointer to a S_usb instance
//! \param  pLun          Pointer to the LUN affected by the command
//! \param  pCommandState Current state of the command
//! \return Operation result code (SUCCESS, ERROR, INCOMPLETE or PARAMETER)
//! \see    S_usb
//! \see    S_lun
//! \see    S_bot_command_state
//------------------------------------------------------------------------------
static unsigned char SBC_Read10(const S_usb         *pUsb,
                                S_lun               *pLun,
                                S_bot_command_state *pCommandState)
{
    unsigned char bStatus;
    unsigned char bResult = BOT_STATUS_INCOMPLETE;
    S_sbc_read_10 *pCommand = (S_sbc_read_10 *) pCommandState->sCbw.pCommand;
    S_bot_transfer *pTransfer = &(pCommandState->sTransfer);

    // Init command state
    if (pCommandState->bState == 0) {

        pCommandState->bState = SBC_STATE_READ;
    }

    // Convert dLength from bytes to blocks
    pCommandState->dLength /= pLun->dBlockSize;

    // Check length
    if (pCommandState->dLength == 0) {

        bResult = BOT_STATUS_SUCCESS;
    }
    else {

        // Command state management
        switch (pCommandState->bState) {
        //------------------
        case SBC_STATE_READ:
        //------------------
            // Read one block of data from the media
            bStatus = LUN_Read(pLun,
                               DWORDB(pCommand->pLogicalBlockAddress),
                               pLun->pReadWriteBuffer,
                               1,
                               (Callback_f) BOT_Callback,
                               (void *) pTransfer);

            // Check operation result code
            if (bStatus != LUN_STATUS_SUCCESS) {

                TRACE_WARNING("W: RBC_Read10: Failed to start reading media\n\r");
                SBC_UpdateSenseData(&(pLun->sRequestSenseData),
                                    SBC_SENSE_KEY_NOT_READY,
                                    SBC_ASC_LOGICAL_UNIT_NOT_READY,
                                    0);
                bResult = BOT_STATUS_ERROR;
            }
            else {

                // Move to next command state
                pCommandState->bState = SBC_STATE_WAIT_READ;
            }
            break;

        //-----------------------
        case SBC_STATE_WAIT_READ:
        //-----------------------
            // Check semaphore value
            if (pTransfer->bSemaphore > 0) {

                TRACE_DEBUG_M("Ok ");

                // Take semaphore and move to next state
                pTransfer->bSemaphore--;
                pCommandState->bState = SBC_STATE_WRITE;
            }
            break;

        //-------------------
        case SBC_STATE_WRITE:
        //-------------------
            // Check the operation result code
            if (pTransfer->bStatus != USB_STATUS_SUCCESS) {

                TRACE_WARNING("W: RBC_Read10: Failed to read media\n\r");
                SBC_UpdateSenseData(&(pLun->sRequestSenseData),
                                    SBC_SENSE_KEY_RECOVERED_ERROR,
                                    SBC_ASC_LOGICAL_BLOCK_ADDRESS_OUT_OF_RANGE,
                                    0);
                bResult = BOT_STATUS_ERROR;
            }
            else {

                // Send the block to the host
                bStatus = USB_Write(pUsb,
                                    BOT_EPT_BULK_IN,
                                    pLun->pReadWriteBuffer,
                                    pLun->dBlockSize,
                                    (Callback_f) BOT_Callback,
                                    (void *) pTransfer);

                // Check operation result code
                if (bStatus != USB_STATUS_SUCCESS) {

                    TRACE_WARNING("W: RBC_Read10: Failed to start to send data\n\r");
                    SBC_UpdateSenseData(&(pLun->sRequestSenseData),
                                        SBC_SENSE_KEY_HARDWARE_ERROR,
                                        0,
                                        0);
                    bResult = BOT_STATUS_ERROR;
                }
                else {

                    TRACE_DEBUG_M("Sending ");

                    // Move to next command state
                    pCommandState->bState = SBC_STATE_WAIT_WRITE;
                }
            }
            break;

        //------------------------
        case SBC_STATE_WAIT_WRITE:
        //------------------------
            // Check semaphore value
            if (pTransfer->bSemaphore > 0) {

                TRACE_DEBUG_M("Sent ");

                // Take semaphore and move to next state
                pTransfer->bSemaphore--;
                pCommandState->bState = SBC_STATE_NEXT_BLOCK;
            }
            break;

        //------------------------------
        case SBC_STATE_NEXT_BLOCK:
        //------------------------------
            // Check operation result code
            if (pTransfer->bStatus != USB_STATUS_SUCCESS) {

                TRACE_WARNING("W: RBC_Read10: Failed to send data\n\r");
                SBC_UpdateSenseData(&(pLun->sRequestSenseData),
                                    SBC_SENSE_KEY_HARDWARE_ERROR,
                                    0,
                                    0);
                bResult = BOT_STATUS_ERROR;
            }
            else {
                TRACE_DEBUG_M("Next ");

                // Update transfer length and block address
                STORE_DWORDB(DWORDB(pCommand->pLogicalBlockAddress) + 1,
                             pCommand->pLogicalBlockAddress);
                pCommandState->dLength--;

                // Check if transfer is finished
                if (pCommandState->dLength == 0) {

                    bResult = BOT_STATUS_SUCCESS;
                }
                else {

                    pCommandState->bState = SBC_STATE_READ;
                }
            }
            break;
        }
    }

    // Convert dLength from blocks to bytes
    pCommandState->dLength *= pLun->dBlockSize;

    return bResult;
}

//------------------------------------------------------------------------------
//! \brief  Performs a READ CAPACITY (10) command.
//!
//!         This function operates asynchronously and must be called multiple
//!         times to complete. A result code of BOT_STATUS_INCOMPLETE indicates
//!         that at least another call of the method is necessary.
//! \param  pUsb          Pointer to a S_usb instance
//! \param  pLun          Pointer to the LUN affected by the command
//! \param  pCommandState Current state of the command
//! \return Operation result code (SUCCESS, ERROR, INCOMPLETE or PARAMETER)
//! \see    S_usb
//! \see    S_lun
//! \see    S_bot_command_state
//------------------------------------------------------------------------------
static unsigned char SBC_ReadCapacity10(const S_usb         *pUsb,
                                        S_lun               *pLun,
                                        S_bot_command_state *pCommandState)
{
    unsigned char bResult = BOT_STATUS_INCOMPLETE;
    unsigned char bStatus;
    S_bot_transfer *pTransfer = &(pCommandState->sTransfer);

    // Initialize command state if needed
    if (pCommandState->bState == 0) {

        pCommandState->bState = SBC_STATE_WRITE;
    }

    // Identify current command state
    switch (pCommandState->bState) {
    //-------------------
    case SBC_STATE_WRITE:
    //-------------------
        // Start the write operation
        bStatus = USB_Write(pUsb,
                            BOT_EPT_BULK_IN,
                            &(pLun->sReadCapacityData),
                            pCommandState->dLength,
                            (Callback_f) BOT_Callback,
                            (void *) pTransfer);

        // Check operation result code
        if (bStatus != USB_STATUS_SUCCESS) {

            TRACE_WARNING("W: RBC_ReadCapacity: Cannot start sending data\n\r");
            bResult = BOT_STATUS_ERROR;
        }
        else {

            // Proceed to next command state
            TRACE_DEBUG_M("Sending ");
            pCommandState->bState = SBC_STATE_WAIT_WRITE;
        }
        break;

    //------------------------
    case SBC_STATE_WAIT_WRITE:
    //------------------------
        // Check semaphore value
        if (pTransfer->bSemaphore > 0) {

            // Take semaphore and terminate command
            pTransfer->bSemaphore--;

            if (pTransfer->bStatus != USB_STATUS_SUCCESS) {

                TRACE_WARNING("W: RBC_ReadCapacity: Cannot send data\n\r");
                bResult = BOT_STATUS_ERROR;
            }
            else {

                TRACE_DEBUG_M("Sent ");
                bResult = BOT_STATUS_SUCCESS;
            }
            pCommandState->dLength -= pTransfer->dBytesTransferred;
        }
        break;
    }

    return bResult;
}

//------------------------------------------------------------------------------
//! \brief  Handles an INQUIRY command.
//!
//!         This function operates asynchronously and must be called multiple
//!         times to complete. A result code of BOT_STATUS_INCOMPLETE indicates
//!         that at least another call of the method is necessary.
//! \param  pUsb          Pointer to a S_usb instance
//! \param  pLun          Pointer to the LUN affected by the command
//! \param  pCommandState Current state of the command
//! \return Operation result code (SUCCESS, ERROR, INCOMPLETE or PARAMETER)
//! \see    S_usb
//! \see    S_lun
//! \see    S_bot_command_state
//------------------------------------------------------------------------------
static unsigned char SBC_Inquiry(const S_usb         *pUsb,
                                 S_lun               *pLun,
                                 S_bot_command_state *pCommandState)
{
    unsigned char  bResult = BOT_STATUS_INCOMPLETE;
    unsigned char  bStatus;
    S_bot_transfer *pTransfer = &(pCommandState->sTransfer);

    // Check if required length is 0
    if (pCommandState->dLength == 0) {

        // Nothing to do
        bResult = BOT_STATUS_SUCCESS;
    }
    // Initialize command state if needed
    else if (pCommandState->bState == 0) {

        pCommandState->bState = SBC_STATE_WRITE;

        // Change additional length field of inquiry data
        pLun->pInquiryData->bAdditionalLength
            = (unsigned char) (pCommandState->dLength - 5);
    }

    // Identify current command state
    switch (pCommandState->bState) {
    //-------------------
    case SBC_STATE_WRITE:
    //-------------------
        // Start write operation
        bStatus = USB_Write(pUsb,
                            BOT_EPT_BULK_IN,
                            (void *) pLun->pInquiryData,
                            pCommandState->dLength,
                            (Callback_f) BOT_Callback,
                            (void *) pTransfer);

        // Check operation result code
        if (bStatus != USB_STATUS_SUCCESS) {

            TRACE_WARNING("W: SPC_Inquiry: Cannot start sending data\n\r");
            bResult = BOT_STATUS_ERROR;
        }
        else {

            // Proceed to next state
            TRACE_DEBUG_M("Sending ");
            pCommandState->bState = SBC_STATE_WAIT_WRITE;
        }
        break;

    //------------------------
    case SBC_STATE_WAIT_WRITE:
    //------------------------
        // Check the semaphore value
        if (pTransfer->bSemaphore > 0) {

            // Take semaphore and terminate command
            pTransfer->bSemaphore--;

            if (pTransfer->bStatus != USB_STATUS_SUCCESS) {

                TRACE_WARNING("W: SPC_Inquiry: Data transfer failed\n\r");
                bResult = BOT_STATUS_ERROR;
            }
            else {

                TRACE_DEBUG_M("Sent ");
                bResult = BOT_STATUS_SUCCESS;
            }

            // Update dLength field
            pCommandState->dLength -= pTransfer->dBytesTransferred;
        }
        break;
    }

    return bResult;
}

//------------------------------------------------------------------------------
//! \brief  Performs a REQUEST SENSE command.
//!
//!         This function operates asynchronously and must be called multiple
//!         times to complete. A result code of BOT_STATUS_INCOMPLETE indicates
//!         that at least another call of the method is necessary.
//! \param  pUsb          Pointer to a S_usb instance
//! \param  pLun          Pointer to the LUN affected by the command
//! \param  pCommandState Current state of the command
//! \return Operation result code (SUCCESS, ERROR, INCOMPLETE or PARAMETER)
//! \see    S_usb
//! \see    S_lun
//! \see    S_bot_command_state
//------------------------------------------------------------------------------
static unsigned char SBC_RequestSense(const S_usb         *pUsb,
                                      S_lun               *pLun,
                                      S_bot_command_state *pCommandState)
{
    unsigned char bResult = BOT_STATUS_INCOMPLETE;
    unsigned char bStatus;
    S_bot_transfer *pTransfer = &(pCommandState->sTransfer);

    // Check if requested length is zero
    if (pCommandState->dLength == 0) {

        // Nothing to do
        bResult = BOT_STATUS_SUCCESS;
    }
    // Initialize command state if needed
    else if (pCommandState->bState == 0) {

        pCommandState->bState = SBC_STATE_WRITE;
    }

    // Identify current command state
    switch (pCommandState->bState) {
    //-------------------
    case SBC_STATE_WRITE:
    //-------------------
        // Start transfer
        bStatus = USB_Write(pUsb,
                            BOT_EPT_BULK_IN,
                            &(pLun->sRequestSenseData),
                            pCommandState->dLength,
                            (Callback_f) BOT_Callback,
                            (void *) pTransfer);

        // Check result code
        if (bStatus != USB_STATUS_SUCCESS) {

            TRACE_WARNING("W: RBC_RequestSense: Cannot start sending data\n\r");
            bResult = BOT_STATUS_ERROR;
        }
        else {

            // Change state
            pCommandState->bState = SBC_STATE_WAIT_WRITE;
        }
        break;

    //------------------------
    case SBC_STATE_WAIT_WRITE:
    //------------------------
        // Check the transfer semaphore
        if (pTransfer->bSemaphore > 0) {

            // Take semaphore and finish command
            pTransfer->bSemaphore--;

            if (pTransfer->bStatus != USB_STATUS_SUCCESS) {

                bResult = BOT_STATUS_ERROR;
            }
            else {

                bResult = BOT_STATUS_SUCCESS;
            }

            // Update pLength
            pCommandState->dLength -= pTransfer->dBytesTransferred;
        }
        break;
    }

    return bResult;
}

//------------------------------------------------------------------------------
//! \brief  Performs a MODE SENSE (6) command.
//!
//!         This function operates asynchronously and must be called multiple
//!         times to complete. A result code of BOT_STATUS_INCOMPLETE indicates
//!         that at least another call of the method is necessary.
//! \param  pUsb          Pointer to a S_usb instance
//! \param  pLun          Pointer to the LUN affected by the command
//! \param  pCommandState Current state of the command
//! \return Operation result code (SUCCESS, ERROR, INCOMPLETE or PARAMETER)
//! \see    S_usb
//! \see    S_lun
//! \see    S_bot_command_state
//------------------------------------------------------------------------------
static unsigned char SBC_ModeSense6(const S_usb         *pUsb,
                                    S_lun               *pLun,
                                    S_bot_command_state *pCommandState)
{
    unsigned char      bResult = BOT_STATUS_INCOMPLETE;
    unsigned char      bStatus;
    S_bot_transfer     *pTransfer = &(pCommandState->sTransfer);

    // Initialize command state if needed
    if (pCommandState->bState == 0) {

        pCommandState->bState = SBC_STATE_WRITE;
    }

    // Check current command state
    switch (pCommandState->bState) {
    //-------------------
    case SBC_STATE_WRITE:
    //-------------------
        // Start transfer
        bStatus = USB_Write(pUsb,
                            BOT_EPT_BULK_IN,
                            (void *) &sModeParameterHeader6,
                            pCommandState->dLength,
                            (Callback_f) BOT_Callback,
                            (void *) pTransfer);

        // Check operation result code
        if (bStatus != USB_STATUS_SUCCESS) {

            TRACE_WARNING("W: SPC_ModeSense6: Cannot start data transfer\n\r");
            bResult = BOT_STATUS_ERROR;
        }
        else {

            // Proceed to next state
            pCommandState->bState = SBC_STATE_WAIT_WRITE;
        }
        break;

    //------------------------
    case SBC_STATE_WAIT_WRITE:
    //------------------------
        TRACE_DEBUG_M("Wait ");

        // Check semaphore value
        if (pTransfer->bSemaphore > 0) {

            // Take semaphore and terminate command
            pTransfer->bSemaphore--;

            if (pTransfer->bStatus != USB_STATUS_SUCCESS) {

                TRACE_WARNING("W: SPC_ModeSense6: Data transfer failed\n\r");
                bResult = BOT_STATUS_ERROR;
            }
            else {

                bResult = BOT_STATUS_SUCCESS;
            }

            // Update dLength field
            pCommandState->dLength -= pTransfer->dBytesTransferred;

        }
        break;
    }

    return bResult;
}

//------------------------------------------------------------------------------
//! \brief  Performs a TEST UNIT READY COMMAND command.
//! \param  pLun          Pointer to the LUN affected by the command
//! \return Operation result code (SUCCESS, ERROR, INCOMPLETE or PARAMETER)
//! \see    S_lun
//------------------------------------------------------------------------------
static unsigned char SBC_TestUnitReady(S_lun *pLun)
{
    unsigned char bResult = BOT_STATUS_ERROR;

    // Check current media state
    switch(pLun->pMedia->bState) {
    //-------------------
    case MED_STATE_READY:
    //-------------------
        // Nothing to do
        TRACE_DEBUG_M("Rdy ");
        bResult = BOT_STATUS_SUCCESS;
        break;

    //------------------
    case MED_STATE_BUSY:
    //------------------
        TRACE_DEBUG_M("Bsy ");
        SBC_UpdateSenseData(&(pLun->sRequestSenseData),
                            SBC_SENSE_KEY_NOT_READY,
                            0,
                            0);
        break;

    //------
    default:
    //------
        TRACE_DEBUG_M("? ");
        SBC_UpdateSenseData(&(pLun->sRequestSenseData),
                            SBC_SENSE_KEY_NOT_READY,
                            SBC_ASC_MEDIUM_NOT_PRESENT,
                            0);
        break;
    }

    return bResult;
}

//------------------------------------------------------------------------------
//      Exported functions
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
//! \brief  Updates the sense data of a LUN with the given key and codes
//! \param  pRequestSenseData             Pointer to the sense data to update
//! \param  bSenseKey                     Sense key
//! \param  bAdditionalSenseCode          Additional sense code
//! \param  bAdditionalSenseCodeQualifier Additional sense code qualifier
//------------------------------------------------------------------------------
void SBC_UpdateSenseData(S_sbc_request_sense_data *pRequestSenseData,
                         unsigned char bSenseKey,
                         unsigned char bAdditionalSenseCode,
                         unsigned char bAdditionalSenseCodeQualifier)
{
    pRequestSenseData->bSenseKey = bSenseKey;
    pRequestSenseData->bAdditionalSenseCode = bAdditionalSenseCode;
    pRequestSenseData->bAdditionalSenseCodeQualifier
        = bAdditionalSenseCodeQualifier;
}

//------------------------------------------------------------------------------
//! \brief  Return information about the transfer length and direction expected
//!         by the device for a particular command.
//! \param  pCommand Pointer to a buffer holding the command to evaluate
//! \param  pLength  Expected length of the data transfer
//! \param  pType    Expected direction of data transfer
//! \param  pLun     Pointer to the LUN affected by the command
//------------------------------------------------------------------------------
bool SBC_GetCommandInformation(void          *pCommand,
                               unsigned int  *pLength,
                               unsigned char *pType,
                               S_lun         *pLun)
{
    S_sbc_command *pSbcCommand = (S_sbc_command *) pCommand;
    bool          isCommandSupported = true;

    // Identify command
    switch (pSbcCommand->bOperationCode) {
    //---------------
    case SBC_INQUIRY:
    //---------------
        (*pType) = BOT_DEVICE_TO_HOST;

        // Allocation length is stored in big-endian format
        (*pLength) = WORDB(pSbcCommand->sInquiry.pAllocationLength);
        break;

    //--------------------
    case SBC_MODE_SENSE_6:
    //--------------------
        (*pType) = BOT_DEVICE_TO_HOST;
        (*pLength) = min(sizeof(S_sbc_mode_parameter_header_6),
                         pSbcCommand->sModeSense6.bAllocationLength);

        // Only "return all pages" command is supported
        if (pSbcCommand->sModeSense6.bPageCode != SBC_PAGE_RETURN_ALL) {

            // Unsupported page
            TRACE_WARNING("W: SBC_GetCommandInformation: Page code not supported (0x%02X)\n\r",
                          pSbcCommand->sModeSense6.bPageCode);
            isCommandSupported = false;
            (*pLength) = 0;
        }
        break;

    //------------------------------------
    case SBC_PREVENT_ALLOW_MEDIUM_REMOVAL:
    //------------------------------------
        (*pType) = BOT_NO_TRANSFER;
        break;

    //---------------------
    case SBC_REQUEST_SENSE:
    //---------------------
        (*pType) = BOT_DEVICE_TO_HOST;
        (*pLength) = pSbcCommand->sRequestSense.bAllocationLength;
        break;

    //-----------------------
    case SBC_TEST_UNIT_READY:
    //-----------------------
        (*pType) = BOT_NO_TRANSFER;
        break;

    //---------------------
    case SBC_READ_CAPACITY_10:
    //---------------------
        (*pType) = BOT_DEVICE_TO_HOST;
        (*pLength) = sizeof(S_sbc_read_capacity_10_data);
        break;

    //---------------
    case SBC_READ_10:
    //---------------
        (*pType) = BOT_DEVICE_TO_HOST;
        (*pLength) = WORDB(pSbcCommand->sRead10.pTransferLength)
                     * pLun->dBlockSize;
        break;

    //----------------
    case SBC_WRITE_10:
    //----------------
        (*pType) = BOT_HOST_TO_DEVICE;
        (*pLength) = WORDB(pSbcCommand->sWrite10.pTransferLength)
                     * pLun->dBlockSize;
        break;

    //-----------------
    case SBC_VERIFY_10:
    //-----------------
        (*pType) = BOT_NO_TRANSFER;
        break;

    //------
    default:
    //------
        isCommandSupported = false;
    }

    // If length is 0, no transfer is expected
    if ((*pLength) == 0) {

        (*pType) = BOT_NO_TRANSFER;
    }

    return isCommandSupported;
}

//------------------------------------------------------------------------------
//! \brief  Processes a SBC command by dispatching it to a subfunction.
//! \param  pUsb          Pointer to a S_usb instance
//! \param  pLun          Pointer to the affected LUN
//! \param  pCommandState Pointer to the current command state
//! \return Operation result code
//------------------------------------------------------------------------------
unsigned char SBC_ProcessCommand(const S_usb         *pUsb,
                                 S_lun               *pLun,
                                 S_bot_command_state *pCommandState)
{
    unsigned char bResult = BOT_STATUS_INCOMPLETE;
    S_sbc_command *pCommand = (S_sbc_command *) pCommandState->sCbw.pCommand;

    // Identify command
    switch (pCommand->bOperationCode) {
    //---------------
    case SBC_READ_10:
    //---------------
        TRACE_DEBUG_M("Read(10) ");

        // Perform the Read10 command
        bResult = SBC_Read10(pUsb,
                             pLun,
                             pCommandState);
        break;

    //----------------
    case SBC_WRITE_10:
    //----------------
        TRACE_DEBUG_M("Write(10) ");

        // Perform the Write10 command
        bResult = SBC_Write10(pUsb,
                              pLun,
                              pCommandState);
        break;

    //---------------------
    case SBC_READ_CAPACITY_10:
    //---------------------
        TRACE_DEBUG_M("RdCapacity(10) ");

        // Perform the ReadCapacity command
        bResult = SBC_ReadCapacity10(pUsb, pLun, pCommandState);
        break;

    //---------------------
    case SBC_VERIFY_10:
    //---------------------
        TRACE_DEBUG_M("Verify(10) ");

        // Nothing to do
        bResult = BOT_STATUS_SUCCESS;
        break;

    //---------------
    case SBC_INQUIRY:
    //---------------
        TRACE_DEBUG_M("Inquiry ");

        // Process Inquiry command
        bResult = SBC_Inquiry(pUsb, pLun, pCommandState);
        break;

    //--------------------
    case SBC_MODE_SENSE_6:
    //--------------------
        TRACE_DEBUG_M("ModeSense(6) ");

        // Process ModeSense6 command
        bResult = SBC_ModeSense6(pUsb, pLun, pCommandState);
        break;

    //-----------------------
    case SBC_TEST_UNIT_READY:
    //-----------------------
        TRACE_DEBUG_M("TstUnitRdy ");

        // Process TestUnitReady command
        bResult = SBC_TestUnitReady(pLun);
        break;

    //---------------------
    case SBC_REQUEST_SENSE:
    //---------------------
        TRACE_DEBUG_M("ReqSense ");

        // Perform the RequestSense command
        bResult = SBC_RequestSense(pUsb, pLun, pCommandState);
        break;

    //------------------------------------
    case SBC_PREVENT_ALLOW_MEDIUM_REMOVAL:
    //------------------------------------
        TRACE_DEBUG_M("PrevAllowRem ");

        // Nothing to do
        bResult = BOT_STATUS_SUCCESS;
        break;

    //------
    default:
    //------
        bResult = BOT_STATUS_PARAMETER;
    }

    return bResult;
}
